/* ******** netUstad: Network Ustadi (Network Master) ******** */
/* This software is Copyright (C) 2004 by Ozkan KIRIK.        */
/* Permission is granted to redistribute and modify this      */
/* software under the terms of the GNU General Public License */
/* available at http://www.gnu.org/                           */
/**************************************************************/


#include "common.h"

#define PID_FILE	"/var/run/netustad.pid"

#define SERVER_STRING	"Server: netustad/0.2-BETA\r\n"

#include "netustad.h"
#include "loadconfig.h"
#include "mimetype.h"
#include "acceptrequest.h"

int		server_sock = -1;
char		server_sid[] = "1234567890123456789012";
char           *default_conf_file;
const char     *VERSION = "v0.2-BETA";


/**********************************************************************/
/* Write message to log file					      */
/**********************************************************************/
void 
log_msg(char *filename, char *msg)
{
	FILE           *log_file;

	log_file = fopen(filename, "a");
	if (!log_file)
		return;
	fprintf(log_file, "%s", msg);
	fclose(log_file);
	return;
}

/**********************************************************************/
/* Catch and parse signals                                            */
/**********************************************************************/
void 
signal_handler(int sig)
{
	time(&currentsec);
	timenow = strtok(ctime(&currentsec), "\n");
	switch (sig) {
	case SIGXFSZ:
		log_msg(mainlogfile, "File size limit exceeded at ");
		log_msg(mainlogfile, timenow);
		log_msg(mainlogfile, "\nProgram Terminated.\n");
		log_msg(mainlogfile, "<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>\n\n");
		if (client_sock != -1)
			close(client_sock);
		shutdown(server_sock, 2);
		close(server_sock);
		exit(1);
		break;
	case SIGPIPE:
		shutdown(client_sock, 2);
		close(client_sock);
		break;
	case SIGHUP:
		log_msg(mainlogfile, "Hangup signal catched at ");
		log_msg(mainlogfile, timenow);
		log_msg(mainlogfile, "\nReloading configuration...\n\n");
		readconfig(default_conf_file);
		log_msg(mainlogfile, "Configuration reloaded successfully.\n\n");
		break;
	case SIGTERM:
		log_msg(mainlogfile, "Terminate signal catched at ");
		log_msg(mainlogfile, timenow);
		log_msg(mainlogfile, "\nClosing active sessions...\n");
		if (client_sock != -1)
			close(client_sock);
		shutdown(server_sock, 1);
		close(server_sock);
		log_msg(mainlogfile, "Bye bye\n");
		log_msg(mainlogfile, "<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>\n\n");
		log_msg(logfile, "<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>\n\n");
		while (waitpid(-1, NULL, WNOHANG) > 0);
		exit(0);
		break;
	}
}


/**************************************/
/* Get the program to become a daemon */
/**************************************/
void 
daemonize()
{
	int		i         ,
			lfp;
	char		str       [10];

	if (getppid() == 1)
		return;		/* already a daemon */
	i = fork();
	if (i < 0)
		exit(1);	/* fork error */
	if (i > 0)
		exit(0);	/* parent exits */
	/* child (daemon) continues */
	setsid();		/* obtain a new process group */
	for (i = getdtablesize(); i >= 0; --i)
		close(i);	/* close all descriptors */

	i = open("/dev/null", O_RDWR);
	dup(i);
	dup(i);			/* handle standart I/O */
	umask(027);		/* set newly created file permissions */
	chdir(workdir);		/* change running directory */

	lfp = open(PID_FILE, O_RDWR | O_CREAT, 0640);
	if (lfp < 0)
		exit(1);	/* can not open */

	if (lockf(lfp, F_TLOCK, 0) < 0) {
		log_msg(mainlogfile, "netUstad is already running!\n");
		log_msg(nameoftty, "netUstad is already running!\n");
		syslog(LOG_WARNING, "netUstad is already running!\n");
		exit(0);
	}			/* can not lock */
	/* first instance continues */
	snprintf(str, sizeof(str), "%d\n", getpid());
	write(lfp, str, strlen(str));	/* record pid to lockfile */
	/* signal(SIGCHLD,SIG_IGN); *//* ignore child */
	signal(SIGTSTP, SIG_IGN);	/* ignore tty signals */
	signal(SIGTTOU, SIG_IGN);
	signal(SIGTTIN, SIG_IGN);
	signal(SIGXFSZ, signal_handler);	/* catch file size limit*/
						/* exceeded sgn */
	signal(SIGPIPE, signal_handler);	/* catch broken pipe sgn */
	signal(SIGHUP, signal_handler);		/* catch hangup signal */
	signal(SIGTERM, signal_handler);	/* catch kill signal */
}

/***************************************************************/
/* Inform the client that a request it has made has a problem. */
/***************************************************************/
void 
bad_request(int client)
{
	char		buf       [128];

	snprintf(buf, sizeof(buf), "HTTP/1.0 400 BAD REQUEST\r\n");
	send(client, buf, sizeof(buf), 0);
	snprintf(buf, sizeof(buf), "Content-type: text/html\r\n");
	send(client, buf, sizeof(buf), 0);
	snprintf(buf, sizeof(buf), "\r\n");
	send(client, buf, sizeof(buf), 0);
	snprintf(buf, sizeof(buf), "<P>Your browser sent a bad request, ");
	send(client, buf, sizeof(buf), 0);
	snprintf(buf, sizeof(buf), "such as a POST without a Content-Length.\r\n");
	send(client, buf, sizeof(buf), 0);
}

/**************************************************/
/* Send client, contents of a file as text format */
/**************************************************/
void 
cat(int client, FILE * resource)
{
	char		buf       [1024];

	fgets(buf, sizeof(buf), resource);
	while (!feof(resource)) {
		send(client, buf, strlen(buf), 0);
		fgets(buf, sizeof(buf), resource);
	}
}

/****************************************************/
/* Send client, contents of a file as binary format */
/****************************************************/
void 
binarycat(int client, FILE * resource)
{
	char		buf;

	while (!feof(resource)) {
		fread(&buf, sizeof(buf), 1, resource);
		send(client, &buf, 1, 0);
	}
}


/**************************************************************/
/* Inform the client that a CGI script could not be executed. */
/**************************************************************/
void 
cannot_execute(int client)
{
	char		buf       [128];

	snprintf(buf, sizeof(buf), "HTTP/1.0 500 Internal Server Error\r\n");
	send(client, buf, strlen(buf), 0);
	snprintf(buf, sizeof(buf), "Content-type: text/html\r\n");
	send(client, buf, strlen(buf), 0);
	snprintf(buf, sizeof(buf), "\r\n");
	send(client, buf, strlen(buf), 0);
	snprintf(buf, sizeof(buf), "<P>Error prohibited CGI execution.\r\n");
	send(client, buf, strlen(buf), 0);
}

/**************************************************/
/* Log the Critical error, and send it to SysLog! */
/**************************************************/
void 
error_die(char *errormsg)
{
	log_msg(mainlogfile, errormsg);
	syslog(LOG_EMERG, errormsg);
	exit(1);
}

/**********************************/
/* Calculate Full Path of Program */
/**********************************/
char           *
calcpath(char *request)
{
	char           *path_env;
	char		buf       [256];
	char		tmp_path  [256];
	char           *env_buf;
	struct stat	st;

	if (strnstr(request, "/", sizeof(request)) == NULL) {
		path_env = getenv("PATH");
		env_buf = strtok(path_env, ":");
		snprintf(tmp_path, sizeof(tmp_path), "%s/netustad", env_buf);
		while (env_buf != NULL) {
			if (stat(tmp_path, &st) != -1) {
				break;
			}
			env_buf = strtok(NULL, ":");
			snprintf(tmp_path, sizeof(tmp_path), "%s/netustad", env_buf);
		}
		snprintf(workdir, sizeof(workdir), "%s", tmp_path);
	} else {
		snprintf(workdir, sizeof(workdir), "%s", realpath(request, buf));
	}
	return workdir;
}



/* This code Derived from: */
/* J. David's webserver */
/* This is a simple webserver. */
/* Created November 1999 by J. David Blackstone. */
/* CSE 4344 (Network concepts), Prof. Zeigler */
/* University of Texas at Arlington */
/* tinyhttpd project: http://tinyhttpd.sourceforge.net */

/**************************/
/* Start of derieved code */
/**************************/

/**********************************************************************/
/*
 * Execute a CGI script.  Will need to set environment variables as
 * appropriate. Parameters: client socket descriptor path to the CGI script
 */
/**********************************************************************/

void 
execute_cgi(int client, const char *path,
	    const char *method, const char *query_string, char *envp[])
{
	char		buf       [1024];
	int		cgi_output [2];
	int		cgi_input  [2];
	pid_t		pid;
	int		status;
	int		i;
	char		c;
	int		numchars = 1;
	int		content_length = -1;

	buf[0] = 'A';
	buf[1] = '\0';
	if (strcasecmp(method, "GET") == 0)
		while ((numchars > 0) && strcmp("\n", buf))	/* read & discard*/
								/* headers */
			numchars = get_line(client, buf, sizeof(buf));
	else {			/* POST */
		numchars = get_line(client, buf, sizeof(buf));
		while ((numchars > 0) && strcmp("\n", buf)) {
			buf[15] = '\0';
			if (strcasecmp(buf, "Content-Length:") == 0)
				content_length = atoi(&(buf[16]));
			numchars = get_line(client, buf, sizeof(buf));
		}
		if (content_length == -1) {
			bad_request(client);
			return;
		}
	}

	snprintf(buf, sizeof(buf), "HTTP/1.0 200 OK\r\n");
	send(client, buf, strlen(buf), 0);
	snprintf(buf, sizeof(buf), "%s", SERVER_STRING);
	send(client, buf, strlen(buf), 0);

	if (pipe(cgi_output) < 0) {
		cannot_execute(client);
		return;
	}
	if (pipe(cgi_input) < 0) {
		cannot_execute(client);
		return;
	}
	if ((pid = fork()) < 0) {
		cannot_execute(client);
		return;
	}
	if (pid == 0) {		/* child: CGI script */
		char		meth_env  [255];
		char		query_env [255];
		char		length_env[255];

		dup2(cgi_output[1], 1);
		dup2(cgi_input[0], 0);
		close(cgi_output[0]);
		close(cgi_input[1]);
		snprintf(meth_env, sizeof(meth_env), "REQUEST_METHOD=%s", method);
		putenv(meth_env);
		if (strcasecmp(method, "GET") == 0) {
			snprintf(query_env, sizeof(query_env), "QUERY_STRING=%s", query_string);
			putenv(query_env);
		} else {	/* POST */
			snprintf(length_env, sizeof(length_env), "CONTENT_LENGTH=%d", content_length);
			putenv(length_env);
		}
		execle(path, path, NULL, envp);
		exit(0);
	} else {		/* parent */
		close(cgi_output[1]);
		close(cgi_input[0]);
		if (strcasecmp(method, "POST") == 0)
			for (i = 0; i < content_length; i++) {
				recv(client, &c, 1, 0);
				write(cgi_input[1], &c, 1);
			}
		while (read(cgi_output[0], &c, 1) > 0)
			send(client, &c, 1, 0);

		close(cgi_output[0]);
		close(cgi_input[1]);
		waitpid(pid, &status, 0);
	}
}

/**********************************************************************/
/*
 * Get a line from a socket, whether the line ends in a newline, carriage
 * return, or a CRLF combination.  Terminates the string read with a null
 * character.  If no newline indicator is found before the end of the buffer,
 * the string is terminated with a null.  If any of the above three line
 * terminators is read, the last character of the string will be a linefeed
 * and the string will be terminated with a null character. Parameters: the
 * socket descriptor the buffer to save the data in the size of the buffer
 * Returns: the number of bytes stored (excluding null)
 */
/**********************************************************************/
int 
get_line(int sock, char *buf, int size)
{
	int		i = 0;
	char		c = '\0';
	int		n;

	while ((i < size - 1) && (c != '\n')) {
		n = recv(sock, &c, 1, 0);
		/* DEBUG printf("%02X\n", c); */
		if (n > 0) {
			if (c == '\r') {
				n = recv(sock, &c, 1, MSG_PEEK);
				/* DEBUG printf("%02X\n", c); */
				if ((n > 0) && (c == '\n'))
					recv(sock, &c, 1, 0);
				else
					c = '\n';
			}
			buf[i] = c;
			i++;
		} else
			c = '\n';
	}
	buf[i] = '\0';

	return (i);
}

/*************************/
/* End of Derieved Code  */
/*************************/



/******************************************/
/* Send client HTTP headers about a file. */
/******************************************/
int 
headers(int client, char *filename)
{
	char		buf       [1024];
	struct stat	filestat;

	strncpy(buf, "HTTP/1.0 200 OK\r\n", sizeof(buf));
	send(client, buf, strlen(buf), 0);
	snprintf(buf, sizeof(buf), "%s", SERVER_STRING);
	send(client, buf, strlen(buf), 0);
	if (strcmp(mimetype(filename), "text/html") != 0 && \
	    strcmp(mimetype(filename), "text/css") != 0) {
		if (stat(filename, &filestat) != 1) {
			snprintf(buf, sizeof(buf), "Accept-Ranges: bytes\r\n");
			send(client, buf, strlen(buf), 0);
			snprintf(buf, sizeof(buf), "Content-Length: %d\r\n", (int)filestat.st_size);
			send(client, buf, strlen(buf), 0);
		}
	}
	snprintf(buf, sizeof(buf), "Connection: close\r\n");
	send(client, buf, strlen(buf), 0);
	snprintf(buf, sizeof(buf), "Content-Type: %s\r\n\r\n", mimetype(filename));
	send(client, buf, strlen(buf), 0);
	return 1;
}

/******************************************/
/* Send a client a 404 not found message. */
/******************************************/
void 
not_found(int client)
{
	char		buf       [1024];

	snprintf(buf, sizeof(buf), "HTTP/1.0 404 NOT FOUND\r\n");
	send(client, buf, strlen(buf), 0);
	snprintf(buf, sizeof(buf), SERVER_STRING);
	send(client, buf, strlen(buf), 0);
	snprintf(buf, sizeof(buf), "Content-Type: text/html\r\n");
	send(client, buf, strlen(buf), 0);
	snprintf(buf, sizeof(buf), "\r\n");
	send(client, buf, strlen(buf), 0);
	snprintf(buf, sizeof(buf), "<HTML><TITLE>Not Found</TITLE>\r\n");
	send(client, buf, strlen(buf), 0);
	snprintf(buf, sizeof(buf), "<BODY>\r\n");
	send(client, buf, strlen(buf), 0);
	snprintf(buf, sizeof(buf), "Your request is unavailable or nonexistant\r\n");
	send(client, buf, strlen(buf), 0);
	snprintf(buf, sizeof(buf), "</BODY></HTML>\r\n");
	send(client, buf, strlen(buf), 0);
}

/********************************************/
/* Open Socket and check errors then listen */
/********************************************/
int 
startup(u_short * port)
{
	int		httpd = 0;
	struct sockaddr_in name;
	int		tmp;

	httpd = socket(PF_INET, SOCK_STREAM, 0);
	if (httpd == -1)
		error_die("Socket error\n");
	memset(&name, 0, sizeof(name));
	tmp = 1;
	setsockopt(httpd, SOL_SOCKET, SO_REUSEADDR, (char *)&tmp, sizeof(tmp));
	name.sin_family = AF_INET;
	name.sin_port = htons(*port);
	name.sin_addr.s_addr = htonl(INADDR_ANY);
	if (bind(httpd, (struct sockaddr *)&name, sizeof(name)) < 0)
		error_die("Bind error\n");
	if (*port == 0) {	/* if dynamically allocating a port */
		int		namelen = sizeof(name);

		if (getsockname(httpd, (struct sockaddr *)&name, &namelen) == -1)
			error_die("Getsockname error\n");
		*port = ntohs(name.sin_port);
	}
	if (listen(httpd, 2) < 0)
		error_die("Listen error\n");
	return (httpd);
}

/**********************************************************************/
/* Inform the client that the requested web method is not implemented */
/**********************************************************************/
void 
unimplemented(int client)
{
	char		buf       [1024];

	snprintf(buf, sizeof(buf), "HTTP/1.0 501 Method Not Implemented\r\n");
	send(client, buf, strlen(buf), 0);
	snprintf(buf, sizeof(buf), SERVER_STRING);
	send(client, buf, strlen(buf), 0);
	snprintf(buf, sizeof(buf), "Content-Type: text/html\r\n");
	send(client, buf, strlen(buf), 0);
	snprintf(buf, sizeof(buf), "\r\n");
	send(client, buf, strlen(buf), 0);
	snprintf(buf, sizeof(buf), "<HTML><HEAD><TITLE>Method Not Implemented\r\n");
	send(client, buf, strlen(buf), 0);
	snprintf(buf, sizeof(buf), "</TITLE></HEAD>\r\n");
	send(client, buf, strlen(buf), 0);
	snprintf(buf, sizeof(buf), "<BODY>HTTP request method not supported.\r\n");
	send(client, buf, strlen(buf), 0);
	snprintf(buf, sizeof(buf), "</BODY></HTML>\r\n");
	send(client, buf, strlen(buf), 0);
}

/**********************************************************************/

int 
main(int argc, char *argv[])
{
	int		client_name_len = sizeof(client_name);
	char		log_msg_text[128];
	char		tmpstr    [128];

	client_sock = -1;
	snprintf(tmpstr, sizeof(tmpstr), "%s", calcpath(argv[0]));
	snprintf(tmpstr, strlen(tmpstr) - 13, "%s", tmpstr);
	snprintf(tmpstr, sizeof(tmpstr), "%s/etc/netustad.conf", tmpstr);
	default_conf_file = strdup(tmpstr);


	/**************************/
	/* Find working directory */
	/**************************/
	/* snprintf(workdir, sizeof(workdir), "%s", calcpath(argv[0])); */
	snprintf(workdir, strlen(workdir) - 12, "%s", workdir);
	snprintf(workdir, sizeof(workdir), "%s/share/netustad/", workdir);


	/*******************/
	/* Check Arguments */
	/*******************/
	if (argc > 1) {		/* If number of parameters more than 0 */
		if (strcmp(argv[1], "-h") == 0 || \
		    strcmp(argv[1], "--help") == 0) {
			printf("\nnetUstad: Network Ustadi (Network Master)\n");
			printf("Copyright (C) 2004   by Ozkan KIRIK\n");
			printf("Usage: netustad [options]\n");
			printf("\n");
			printf("Options:\n");
			printf("\t-h\tShow this help screen\n");
			printf("\t-v\tShow version\n");
			printf("\t-c\tUse following parameter as configuration file");
			printf("\n");
			exit(0);
		} else if (strcmp(argv[1], "-v") == 0 || \
			   strcmp(argv[1], "--version") == 0) {
			printf("\nnetUstad: Network Ustadi (Network Master)\n");
			printf("Copyright (C) 2004  by Ozkan KIRIK\n\n");
			printf("Version: %s\n\n", VERSION);
			exit(0);
		} else if (strcmp(argv[1], "-c") == 0) {
			if (argc == 3) {
				default_conf_file = strdup(argv[2]);
			} else {
				printf("\nnetUstad: Invalid Number Of Arguments\n\n");
				exit(1);
			}
		} else {
			printf("\nnetUstad: Invalid Argument\n\n");
			exit(1);
		}
	}
	/**********************/
	/* Start Main Program */
	/**********************/

	readconfig(default_conf_file);
	nameoftty = ttyname(0);
	daemonize();
	server_sock = startup(&port);	/* Open Socket & Listen */

	time(&currentsec);
	log_msg(mainlogfile, "<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>\n");
	snprintf(log_msg_text, sizeof(log_msg_text), "netUstad is started at %s", strtok(ctime(&currentsec), "\n"));
	log_msg(mainlogfile, log_msg_text);
	snprintf(log_msg_text, sizeof(log_msg_text), "\nListening port %d\n", port);
	log_msg(mainlogfile, log_msg_text);
	log_msg(mainlogfile, "Ready for requests...\n\n");
	snprintf(log_msg_text, sizeof(log_msg_text), "\nnetUstad is started\nListening port %d\n", port);
	log_msg(nameoftty, log_msg_text);

	while (1) {
		client_sock = accept(server_sock,
				     (struct sockaddr *)&client_name,
				     &client_name_len);

		if (client_sock == -1)
			continue;
		accept_request(client_sock);
		close(client_sock);
	}

	close(server_sock);

	return (0);
}
